@@ -34,8 +34,8 @@ use crate::decimal128::Decimal128;
 use crate::decimal32::Decimal32;
 use crate::decimal64::Decimal64;
 use crate::error::{
-    InvalidCoefficientError, InvalidExponentError, InvalidPrecisionError, ParseDecimalError,
-    TryFromDecimalError, TryIntoDecimalError,
+    FromBcdError, InvalidCoefficientError, InvalidExponentError, InvalidPrecisionError,
+    ParseDecimalError, TryFromDecimalError, TryIntoDecimalError,
 };
 
 fn validate_n(n: usize) {
@@ -55,8 +55,11 @@ fn validate_n(n: usize) {
 ///
 /// `N` must be at least 12 and no greater than 999,999,999, though typically
 /// the stack size implies a smaller maximum for `N`. Due to limitations with
-/// constant generics it is not yet possible to enforce these restrictions
-/// at compile time, so they are checked at runtime.
+/// constant generics it is not yet possible to enforce these restrictions at
+/// compile time, so they are checked at runtime.
+///
+/// For more details about e.g. the struct's fields, see the [upstream
+/// documentation](http://speleotrove.com/decimal/dnnumb.html).
 #[repr(C)]
 #[derive(Clone, Copy)]
 #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
@@ -202,8 +205,7 @@ impl<const N: usize> Decimal<N> {
     /// number of digits.
     ///
     /// This function is public and accepts a `u32` instead of a `Decimal` to
-    /// aid in decomposing ([`Self::to_raw_parts`]) and recomposing
-    /// ([`Self::from_raw_parts`]) values.
+    /// aid in recomposing ([`Self::from_raw_parts`]) values.
     pub fn digits_to_lsu_elements_len(digits: u32) -> usize {
         (usize::try_from(digits).unwrap() + decnumber_sys::DECDPUN - 1) / decnumber_sys::DECDPUN
     }
@@ -318,57 +320,35 @@ impl<const N: usize> Decimal<N> {
     /// converted to `u8`.
     ///
     /// The meaning of these parts are unspecified and subject to change.
-    pub fn to_raw_parts(&self) -> (u32, i32, u8, &[u8]) {
-        // SAFETY: `lsu` (returned by `coefficient_units()`) is a `&[u16]`, so
-        // each element can safely be transmuted into two `u8`s.
-        let (prefix, lsu, suffix) = unsafe { self.coefficient_units().align_to::<u8>() };
-        // The `u8` aligned version of the `lsu` should have twice as many
-        // elements as we expect for the `u16` version.
-        assert!(
-            lsu.len() == Self::digits_to_lsu_elements_len(self.digits) * 2,
-            "u8 version of LSU contained the wrong number of elements; expected {}, but got {}",
-            Self::digits_to_lsu_elements_len(self.digits) * 2,
-            lsu.len()
-        );
-        // There should be no unaligned elements in the prefix or suffix.
-        assert!(prefix.is_empty() && suffix.is_empty());
-        (self.digits, self.exponent, self.bits, lsu)
+    pub fn to_raw_parts(&self) -> (u32, i32, u8, &[u16]) {
+        (
+            self.digits,
+            self.exponent,
+            self.bits,
+            self.coefficient_units(),
+        )
     }
 
     /// Returns a `Decimal::<N>` with the supplied raw parts, which should be
-    /// generated using [`Decimal::to_raw_parts`].
-    ///
-    /// # Safety
-    ///
-    /// The raw parts must be valid according to the guarantees required by the
-    /// underlying C library, or undefined behavior can result. The easiest way
-    /// to uphold these guarantees is to ensure the raw parts originate from a
-    /// call to `Decimal::to_raw_parts`.
+    /// generated using [`Decimal::to_raw_parts`] on a machine with compatible
+    /// architecture.
     ///
     /// # Panics
     ///
-    /// If `lsu_u8` is not a slice that can be recomposed into a `&[16]` with
-    /// the number of digits implicitly specified by the `digits` parameter,
-    /// i.e. essentially `ceil(digits / decnumber_sys::DECDPUN)`. You can determine the appropriate
-    /// number of elements in `lsu_u8` using 2 *
-    /// [`Decimal::digits_to_lsu_elements_len`].
-    pub unsafe fn from_raw_parts(digits: u32, exponent: i32, bits: u8, lsu_u8: &[u8]) -> Self {
+    /// If `lsu_in` is not a slice with the number of digits implicitly
+    /// specified by the `digits` parameter, i.e. essentially `ceil(digits /
+    /// decnumber_sys::DECDPUN)`.
+    pub fn from_raw_parts(digits: u32, exponent: i32, bits: u8, lsu_in: &[u16]) -> Self {
+        let lsu_expected_len = Self::digits_to_lsu_elements_len(digits);
         assert!(
-            lsu_u8.len() == Self::digits_to_lsu_elements_len(digits) * 2,
-            "Expected lsu_u8 to have {} elements, but instead got {}",
-            Self::digits_to_lsu_elements_len(digits) * 2,
-            lsu_u8.len()
+            lsu_in.len() == lsu_expected_len,
+            "Expected lsu_in to have {} elements, but instead got {}",
+            lsu_expected_len,
+            lsu_in.len()
         );
 
         let mut lsu = [0; N];
-        for (i, v) in lsu_u8.chunks(2).enumerate() {
-            // Use native endian ordering, which matches the ordering generated
-            // for the `lsu`/`coefficient_units()` in `to_raw_parts`.
-            lsu[i] = u16::from_ne_bytes(
-                (*v).try_into()
-                    .expect("converting slice of 2 u8s to [u8; 2] to succeed"),
-            );
-        }
+        lsu[0..lsu_expected_len].copy_from_slice(&lsu_in);
 
         Decimal::<N> {
             digits,
@@ -378,10 +358,120 @@ impl<const N: usize> Decimal<N> {
         }
     }
 
+    /// Returns `self` as a [Packed Decimal][pd] number, including its scale
+    /// (i.e. its negated exponent) or `None` for special values.
+    ///
+    /// [pd]: http://speleotrove.com/decimal/dnpack.html
+    pub fn to_packed_bcd(&mut self) -> Option<(Vec<u8>, i32)> {
+        if self.is_special() {
+            return None;
+        }
+        let mut len = (usize::try_from(self.digits).unwrap() + 1) / 2;
+        if self.digits % 2 == 0 {
+            // Ensure space for sign nibble
+            len += 1;
+        }
+
+        let mut bcd = vec![0; len];
+        let mut scale: i32 = 0;
+        let ret = unsafe {
+            decnumber_sys::decPackedFromNumber(
+                bcd.as_mut_ptr() as *mut u8,
+                len.try_into().unwrap(),
+                &mut scale as *mut i32,
+                self.as_mut_ptr() as *mut decnumber_sys::decNumber,
+            )
+        };
+        // Null returned only for special values (already handled) or if `self`
+        // didn't fit within `bcd` (shouldn't happen).
+        assert!(!ret.is_null());
+        Some((bcd, scale))
+    }
+
+    /// Takes [Packed Decimal][pd] values and their scales (generated by
+    /// [`Self::to_packed_bcd`]) and returns a `Decimal`.
+    ///
+    /// # Errors
+    ///
+    /// - `bcd` contains more digits than the coefficient permits
+    /// - The adjusted exponent is out of range
+    /// - No sign nibble was found
+    /// - A sign nibble was found before the final nibble
+    ///
+    /// [pd]: http://speleotrove.com/decimal/dnpack.html
+    pub fn from_packed_bcd(bcd: &[u8], scale: i32) -> Result<Decimal<N>, FromBcdError> {
+        let mut d = Decimal::default();
+        let ret = unsafe {
+            decnumber_sys::decPackedToNumber(
+                bcd.as_ptr() as *const u8,
+                bcd.len().try_into().unwrap(),
+                &scale as *const i32,
+                d.as_mut_ptr() as *mut decnumber_sys::decNumber,
+            )
+        };
+
+        if ret.is_null() {
+            Err(FromBcdError)
+        } else {
+            Ok(d)
+        }
+    }
+
     /// Returns a string of the number in standard notation, i.e. guaranteed to
     /// not be scientific notation.
     pub fn to_standard_notation_string(&self) -> String {
-        to_standard_notation_string!(self)
+        if !self.is_finite() {
+            return self.to_string();
+        }
+        let digits = self.coefficient_digits();
+        let digits = {
+            let i = digits
+                .iter()
+                .position(|d| *d != 0)
+                .unwrap_or(digits.len() - 1);
+            &digits[i..]
+        };
+        let ndigits = digits.len() as i32;
+        let e = self.exponent();
+        // We allocate space for all the digits plus a possible "-0." prefix.
+        // This is usually an overestimate but is an underestimate for very
+        // large or very small scales.
+        let mut out = String::with_capacity(digits.len() + 3);
+        if self.is_negative() {
+            out.push('-');
+        }
+
+        if e >= 0 {
+            // All digits before the decimal point.
+            for d in digits {
+                out.push(char::from(b'0' + *d));
+            }
+            if !self.is_zero() {
+                for _ in 0..e {
+                    out.push('0');
+                }
+            }
+        } else if ndigits > -e {
+            // Digits span the decimal point.
+            let e = (ndigits + e) as usize;
+            for d in &digits[..e] {
+                out.push(char::from(b'0' + *d));
+            }
+            out.push('.');
+            for d in &digits[e..] {
+                out.push(char::from(b'0' + *d));
+            }
+        } else {
+            // All digits after the decimal point.
+            out.push_str("0.");
+            for _ in 0..(-e - ndigits) {
+                out.push('0');
+            }
+            for d in digits {
+                out.push(char::from(b'0' + *d));
+            }
+        }
+        out
     }
 
     /// Removes insignificant trailing zeros from a number, unconditionally, and
@@ -598,6 +688,11 @@ impl<const N: usize> FromStr for Decimal<N> {
 /// by `$d`) into primitive integers (`$p`).
 macro_rules! __decnum_tryinto_primitive {
     ($p:ty, $cx:expr, $max_digits:literal, $d:expr, $allow_neg:expr) => {{
+        $cx.rescale(&mut $d, &Decimal::<N>::zero());
+
+        // inexact indicates you rounded away non-zero digits during rescale.
+        let inexact = $cx.status().inexact();
+
         let mut fail = || -> TryFromDecimalError {
             let mut s = $cx.status();
             s.set_invalid_operation();
@@ -607,8 +702,8 @@ macro_rules! __decnum_tryinto_primitive {
 
         if $d.is_special()
             || $d.digits() > $max_digits
-            || $d.exponent() != 0
             || (!$allow_neg && $d.is_negative())
+            || inexact
         {
             return Err(fail());
         }
@@ -647,6 +742,46 @@ macro_rules! decnum_tryinto_primitive_uint {
     }};
 }
 
+/// Refer to the comments on [`Context<Decimal<N>>::try_into_i32()`], which also apply to
+/// this trait.
+impl<const N: usize> TryFrom<Decimal<N>> for i8 {
+    type Error = TryFromDecimalError;
+    fn try_from(n: Decimal<N>) -> Result<i8, Self::Error> {
+        let mut cx = Context::<Decimal<N>>::default();
+        cx.try_into_i8(n)
+    }
+}
+
+/// Refer to the comments on [`Context<Decimal<N>>::try_into_i32()`], which also apply to
+/// this trait.
+impl<const N: usize> TryFrom<Decimal<N>> for u8 {
+    type Error = TryFromDecimalError;
+    fn try_from(n: Decimal<N>) -> Result<u8, Self::Error> {
+        let mut cx = Context::<Decimal<N>>::default();
+        cx.try_into_u8(n)
+    }
+}
+
+/// Refer to the comments on [`Context<Decimal<N>>::try_into_i32()`], which also apply to
+/// this trait.
+impl<const N: usize> TryFrom<Decimal<N>> for i16 {
+    type Error = TryFromDecimalError;
+    fn try_from(n: Decimal<N>) -> Result<i16, Self::Error> {
+        let mut cx = Context::<Decimal<N>>::default();
+        cx.try_into_i16(n)
+    }
+}
+
+/// Refer to the comments on [`Context<Decimal<N>>::try_into_i32()`], which also apply to
+/// this trait.
+impl<const N: usize> TryFrom<Decimal<N>> for u16 {
+    type Error = TryFromDecimalError;
+    fn try_from(n: Decimal<N>) -> Result<u16, Self::Error> {
+        let mut cx = Context::<Decimal<N>>::default();
+        cx.try_into_u16(n)
+    }
+}
+
 /// Refer to the comments on [`Context<Decimal<N>>::try_into_i32()`], which also apply to
 /// this trait.
 impl<const N: usize> TryFrom<Decimal<N>> for i32 {
@@ -800,6 +935,30 @@ impl<const N: usize> From<f64> for Decimal<N> {
     }
 }
 
+impl<const N: usize> From<i8> for Decimal<N> {
+    fn from(n: i8) -> Decimal<N> {
+        Decimal::<N>::from(i32::from(n))
+    }
+}
+
+impl<const N: usize> From<u8> for Decimal<N> {
+    fn from(n: u8) -> Decimal<N> {
+        Decimal::<N>::from(u32::from(n))
+    }
+}
+
+impl<const N: usize> From<i16> for Decimal<N> {
+    fn from(n: i16) -> Decimal<N> {
+        Decimal::<N>::from(i32::from(n))
+    }
+}
+
+impl<const N: usize> From<u16> for Decimal<N> {
+    fn from(n: u16) -> Decimal<N> {
+        Decimal::<N>::from(u32::from(n))
+    }
+}
+
 impl<const N: usize> From<i32> for Decimal<N> {
     fn from(n: i32) -> Decimal<N> {
         let mut d = Decimal::default();
@@ -967,7 +1126,7 @@ impl<const N: usize> Context<Decimal<N>> {
 
     /// Sets the context's precision.
     ///
-    /// The precision must be greater than one and no greater than `N * 3`.
+    /// The precision must be at least one and no greater than `N * 3`.
     pub fn set_precision(&mut self, precision: usize) -> Result<(), InvalidPrecisionError> {
         if precision < 1 || precision > N * decnumber_sys::DECDPUN {
             return Err(InvalidPrecisionError);
@@ -1228,20 +1387,61 @@ impl<const N: usize> Context<Decimal<N>> {
         decimal_from_unsigned_int!(self, n)
     }
 
+    /// Attempts to convert `d` to `u8` or fails if not possible.
+    ///
+    /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
+    /// function.
+    pub fn try_into_u8(&mut self, d: Decimal<N>) -> Result<u8, TryFromDecimalError> {
+        let i = self.try_into_u32(d)?;
+        u8::try_from(i).map_err(|_| TryFromDecimalError)
+    }
+
+    /// Attempts to convert `d` to `i8` or fails if not possible.
+    ///
+    /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
+    /// function.
+    pub fn try_into_i8(&mut self, d: Decimal<N>) -> Result<i8, TryFromDecimalError> {
+        let i = self.try_into_i32(d)?;
+        i8::try_from(i).map_err(|_| TryFromDecimalError)
+    }
+
+    /// Attempts to convert `d` to `u16` or fails if not possible.
+    ///
+    /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
+    /// function.
+    pub fn try_into_u16(&mut self, d: Decimal<N>) -> Result<u16, TryFromDecimalError> {
+        let i = self.try_into_u32(d)?;
+        u16::try_from(i).map_err(|_| TryFromDecimalError)
+    }
+
+    /// Attempts to convert `d` to `i16` or fails if not possible.
+    ///
+    /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
+    /// function.
+    pub fn try_into_i16(&mut self, d: Decimal<N>) -> Result<i16, TryFromDecimalError> {
+        let i = self.try_into_i32(d)?;
+        i16::try_from(i).map_err(|_| TryFromDecimalError)
+    }
+
     /// Attempts to convert `d` to `i32` or fails if not possible. Note that
     /// when returning an error, `self`'s [`context::Status`] is set to
-    /// `invalid_operation`.
+    /// `invalid_operation` in addition to using Rust's `Err` return value.
     ///
     /// Note that this function:
+    /// - Accepts any value that can be rescaled to an exponent of 0 without
+    ///   becoming inexact. For example, `123.000` and `123E2` are valid
+    ///   `Decimal` values.
+    ///
+    ///   The corollary is that values that cannot be rescaled to an exponent of
+    ///   0 error.
     /// - Errors if `self.status()` is set to `invalid_operation` irrespective
     ///   of whether or not this specific invocation of the function set that
     ///   status.
-    /// - Fails for values that are representable as integers but whose
-    ///   exponents are > 0. Consider calling `Context::quantize` to set the
-    ///   exponent to 0 before calling this function.
-    pub fn try_into_i32(&mut self, d: Decimal<N>) -> Result<i32, TryFromDecimalError> {
+    pub fn try_into_i32(&mut self, mut d: Decimal<N>) -> Result<i32, TryFromDecimalError> {
+        self.rescale(&mut d, &Decimal::<N>::zero());
         let i = unsafe { decnumber_sys::decNumberToInt32(d.as_ptr(), &mut self.inner) };
-        if self.status().invalid_operation() {
+        // inexact indicates you rounded away non-zero digits during rescale.
+        if self.status().invalid_operation() || self.status().inexact() {
             Err(TryFromDecimalError)
         } else {
             Ok(i)
@@ -1252,9 +1452,11 @@ impl<const N: usize> Context<Decimal<N>> {
     ///
     /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
     /// function.
-    pub fn try_into_u32(&mut self, d: Decimal<N>) -> Result<u32, TryFromDecimalError> {
+    pub fn try_into_u32(&mut self, mut d: Decimal<N>) -> Result<u32, TryFromDecimalError> {
+        self.rescale(&mut d, &Decimal::<N>::zero());
         let i = unsafe { decnumber_sys::decNumberToUInt32(d.as_ptr(), &mut self.inner) };
-        if self.status().invalid_operation() {
+        // inexact indicates you rounded away non-zero digits during rescale.
+        if self.status().invalid_operation() || self.status().inexact() {
             Err(TryFromDecimalError)
         } else {
             Ok(i)
@@ -1267,7 +1469,7 @@ impl<const N: usize> Context<Decimal<N>> {
     /// function.
     #[cfg(target_pointer_width = "32")]
     pub fn try_into_isize(&mut self, d: Decimal<N>) -> Result<isize, TryFromDecimalError> {
-        let d = self.try_into_i32(n)?;
+        let d = self.try_into_i32(d)?;
         Ok(d as isize)
     }
 
@@ -1285,7 +1487,7 @@ impl<const N: usize> Context<Decimal<N>> {
     ///
     /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
     /// function.
-    pub fn try_into_i64(&mut self, d: Decimal<N>) -> Result<i64, TryFromDecimalError> {
+    pub fn try_into_i64(&mut self, mut d: Decimal<N>) -> Result<i64, TryFromDecimalError> {
         decnum_tryinto_primitive_int!(i64, self, 19, d)
     }
 
@@ -1293,7 +1495,7 @@ impl<const N: usize> Context<Decimal<N>> {
     ///
     /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
     /// function.
-    pub fn try_into_i128(&mut self, d: Decimal<N>) -> Result<i128, TryFromDecimalError> {
+    pub fn try_into_i128(&mut self, mut d: Decimal<N>) -> Result<i128, TryFromDecimalError> {
         decnum_tryinto_primitive_int!(i128, self, 39, d)
     }
 
@@ -1303,7 +1505,7 @@ impl<const N: usize> Context<Decimal<N>> {
     /// function.
     #[cfg(target_pointer_width = "32")]
     pub fn try_into_usize(&mut self, d: Decimal<N>) -> Result<usize, TryFromDecimalError> {
-        let d = self.try_into_u32(n)?;
+        let d = self.try_into_u32(d)?;
         Ok(d as usize)
     }
 
@@ -1321,7 +1523,7 @@ impl<const N: usize> Context<Decimal<N>> {
     ///
     /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
     /// function.
-    pub fn try_into_u64(&mut self, d: Decimal<N>) -> Result<u64, TryFromDecimalError> {
+    pub fn try_into_u64(&mut self, mut d: Decimal<N>) -> Result<u64, TryFromDecimalError> {
         decnum_tryinto_primitive_uint!(u64, self, 20, d)
     }
 
@@ -1329,7 +1531,7 @@ impl<const N: usize> Context<Decimal<N>> {
     ///
     /// Refer to the comments on [`Self::try_into_i32()`], which also apply to this
     /// function.
-    pub fn try_into_u128(&mut self, d: Decimal<N>) -> Result<u128, TryFromDecimalError> {
+    pub fn try_into_u128(&mut self, mut d: Decimal<N>) -> Result<u128, TryFromDecimalError> {
         decnum_tryinto_primitive_uint!(u128, self, 39, d)
     }
 
@@ -1641,7 +1843,7 @@ impl<const N: usize> Context<Decimal<N>> {
         }
     }
 
-    /// Rescales `n` to have an exponent of `exp`.
+    /// Rescales `lhs` to have an exponent of `rhs`.
     pub fn rescale<const M: usize>(&mut self, lhs: &mut Decimal<N>, rhs: &Decimal<M>) {
         unsafe {
             decnumber_sys::decNumberRescale(
@@ -1661,6 +1863,38 @@ impl<const N: usize> Context<Decimal<N>> {
         }
     }
 
+    /// Rounds `n` at a given "place from the left" in the number, akin to a
+    /// shift right, round, and shift left.
+    ///
+    /// Note that this rounding will not drop integral digits (i.e those
+    /// representing values at least 1), but can round off fractional values.
+    ///
+    /// `place` must be at least one and no greater than `N * 3`, i.e. a valid
+    /// precision.
+    pub fn round_to_place(
+        &mut self,
+        n: &mut Decimal<N>,
+        place: usize,
+    ) -> Result<(), InvalidPrecisionError> {
+        let precision = self.precision();
+        self.set_precision(place)?;
+        self.plus(n);
+        self.set_precision(precision)
+    }
+
+    /// Identical to [`round_to_place`] but simultaneously performs a [`reduce`]
+    /// operation, as well.
+    pub fn round_reduce_to_place(
+        &mut self,
+        n: &mut Decimal<N>,
+        place: usize,
+    ) -> Result<(), InvalidPrecisionError> {
+        let precision = self.precision();
+        self.set_precision(place)?;
+        self.reduce(n);
+        self.set_precision(precision)
+    }
+
     /// Shifts the digits of `lhs` by `rhs`, storing the result in `lhs`.
     ///
     /// If `rhs` is positive, shifts to the left. If `rhs` is negative, shifts
