@@ -840,7 +840,13 @@ fn test_standard_notation_dec_64() {
         for t in tests {
             cx.set_exponent(&mut d, t.0);
             assert_eq!(t.1, d.to_string());
-            assert_eq!(t.2, d.to_standard_notation_string());
+            assert_eq!(
+                t.2,
+                d.to_standard_notation_string(),
+                "{} should render {} as its standard notation string",
+                d,
+                t.2
+            );
         }
     }
 
@@ -1500,6 +1506,7 @@ fn test_cx_into_i64() {
     inner(i64::MAX / 3);
     inner(i64::MIN / 7);
     inner(i64::MAX / 7);
+    inner(1627418240000);
 }
 
 #[test]
@@ -1513,6 +1520,7 @@ fn test_cx_into_u64() {
     inner(u64::MAX);
     inner(u64::MAX / 3);
     inner(u64::MAX / 7);
+    inner(1627418240000);
 }
 
 #[test]
@@ -1528,6 +1536,7 @@ fn test_cx_into_i128() {
     inner(i128::MAX / 3);
     inner(i128::MIN / 7);
     inner(i128::MAX / 7);
+    inner(1627418240000);
 }
 
 #[test]
@@ -1542,12 +1551,21 @@ fn test_cx_into_u128() {
     inner(u128::MAX / 3);
     inner(u128::MIN / 7);
     inner(u128::MAX / 7);
+    inner(1627418240000);
 }
 
 #[test]
 fn test_decnum_tryinto_primitive() {
     const WIDTH: usize = 14;
     let mut cx = Context::<Decimal<WIDTH>>::default();
+    let min_u8 = Decimal::<WIDTH>::from(u8::MIN);
+    let max_u8 = Decimal::<WIDTH>::from(u8::MAX);
+    let min_i8 = Decimal::<WIDTH>::from(i8::MIN);
+    let max_i8 = Decimal::<WIDTH>::from(i8::MAX);
+    let min_u16 = Decimal::<WIDTH>::from(u16::MIN);
+    let max_u16 = Decimal::<WIDTH>::from(u16::MAX);
+    let min_i16 = Decimal::<WIDTH>::from(i16::MIN);
+    let max_i16 = Decimal::<WIDTH>::from(i16::MAX);
     let min_u32 = Decimal::<WIDTH>::from(u32::MIN);
     let max_u32 = Decimal::<WIDTH>::from(u32::MAX);
     let min_i32 = Decimal::<WIDTH>::from(i32::MIN);
@@ -1560,8 +1578,15 @@ fn test_decnum_tryinto_primitive() {
     let max_u128 = cx.from_u128(u128::MAX);
     let min_i128 = cx.from_i128(i128::MIN);
     let max_i128 = cx.from_i128(i128::MAX);
+    let trailing_zeroes_int = Decimal::<WIDTH>::from(1927418240000i64);
+    let trailing_zeroes_frac = cx.parse("12345.000").unwrap();
+    let scientific_notation = cx.parse("2E5").unwrap();
 
     let inner = |v: &Decimal<WIDTH>| {
+        let i_u8 = u8::try_from(*v);
+        let i_i8 = i8::try_from(*v);
+        let i_u16 = u16::try_from(*v);
+        let i_i16 = i16::try_from(*v);
         let i_u32 = u32::try_from(*v);
         let i_i32 = i32::try_from(*v);
         let i_u64 = u64::try_from(*v);
@@ -1569,6 +1594,34 @@ fn test_decnum_tryinto_primitive() {
         let i_u128 = u128::try_from(*v);
         let i_i128 = i128::try_from(*v);
 
+        // u8
+        if v >= &min_u8 && v <= &max_u8 {
+            assert!(i_u8.is_ok());
+        } else {
+            assert!(i_u8.is_err());
+        }
+
+        // i8
+        if v >= &min_i8 && v <= &max_i8 {
+            assert!(i_i8.is_ok());
+        } else {
+            assert!(i_i8.is_err());
+        }
+
+        // u16
+        if v >= &min_u16 && v <= &max_u16 {
+            assert!(i_u16.is_ok());
+        } else {
+            assert!(i_u16.is_err());
+        }
+
+        // i16
+        if v >= &min_i16 && v <= &max_i16 {
+            assert!(i_i16.is_ok());
+        } else {
+            assert!(i_i16.is_err());
+        }
+
         // u32
         if v >= &min_u32 && v <= &max_u32 {
             assert!(i_u32.is_ok());
@@ -1624,6 +1677,9 @@ fn test_decnum_tryinto_primitive() {
     inner(&max_u128);
     inner(&min_i128);
     inner(&max_i128);
+    inner(&trailing_zeroes_int);
+    inner(&trailing_zeroes_frac);
+    inner(&scientific_notation);
 
     fn inner_expect_failure(s: &str) {
         let mut cx = Context::<Decimal<WIDTH>>::default();
@@ -1640,43 +1696,11 @@ fn test_decnum_tryinto_primitive() {
     inner_expect_failure("-1.2");
     inner_expect_failure("1E-2");
     inner_expect_failure("-1E-2");
-    inner_expect_failure("2E10");
     inner_expect_failure("2E-10");
-    inner_expect_failure("9E40");
     inner_expect_failure("Infinity");
     inner_expect_failure("-Infinity");
     inner_expect_failure("NaN");
-
-    /// Values with non-zero exponents cannot be converted to primitive ints,
-    /// but quantizing the values to an exponent of 0 can convert them to valid
-    /// integer representations.
-    fn quantizable(s: i32, e: i32, valid: bool) {
-        let mut cx = Context::<Decimal<WIDTH>>::default();
-        let mut v = Decimal::<WIDTH>::from(s);
-        v.set_exponent(e);
-        assert!(u32::try_from(v).is_err());
-        assert!(i32::try_from(v).is_err());
-        assert!(u64::try_from(v).is_err());
-        assert!(i64::try_from(v).is_err());
-        assert!(u128::try_from(v).is_err());
-        assert!(i128::try_from(v).is_err());
-        cx.quantize(&mut v, &Decimal::<WIDTH>::zero());
-        if valid {
-            assert!(u32::try_from(v).is_ok());
-            assert!(i32::try_from(v).is_ok());
-            assert!(u64::try_from(v).is_ok());
-            assert!(i64::try_from(v).is_ok());
-            assert!(u128::try_from(v).is_ok());
-            assert!(i128::try_from(v).is_ok());
-        } else {
-            assert!(cx.status().invalid_operation());
-        }
-    }
-
-    quantizable(0, 2, true);
-    quantizable(1, 2, true);
-    quantizable(1, 43, false);
-    quantizable(9999, 99999999, false);
+    inner_expect_failure("123E41");
 }
 
 #[test]
@@ -1929,7 +1953,7 @@ fn decnum_raw_parts() {
         let mut cx = Context::<Decimal<N>>::default();
         let d = cx.parse(s).unwrap();
         let (digits, exponent, bits, lsu) = d.to_raw_parts();
-        let r = unsafe { Decimal::<N>::from_raw_parts(digits, exponent, bits, &lsu) };
+        let r = Decimal::<N>::from_raw_parts(digits, exponent, bits, &lsu);
         if d.is_nan() {
             assert!(r.is_nan())
         } else {
@@ -1953,3 +1977,83 @@ fn decnum_raw_parts() {
     inner("-Infinity", None);
     inner("NaN", None);
 }
+
+#[test]
+fn decnum_raw_parts_bcd() {
+    fn inner(s: &str, o: Option<i128>) {
+        const N: usize = 13;
+        let mut cx = Context::<Decimal<N>>::default();
+        let mut d = cx.parse(s).unwrap();
+        let ret = d.to_packed_bcd();
+        if d.is_special() {
+            assert!(ret.is_none())
+        } else {
+            let (bcd, scale) = ret.unwrap();
+            let r = Decimal::<N>::from_packed_bcd(&bcd, scale).unwrap();
+            assert_eq!(d, r);
+            if let Some(o) = o {
+                let o = cx.from_i128(o);
+                assert_eq!(o, r);
+            }
+        }
+    }
+    inner("1", Some(1));
+    inner("-1", Some(-1));
+    inner("0.00", Some(0));
+    inner("987654321", Some(987654321));
+    inner("-987654321", Some(-987654321));
+    inner(&i128::MAX.to_string(), Some(i128::MAX));
+    inner(&i128::MIN.to_string(), Some(i128::MIN));
+    inner("98765.4321", None);
+    inner("-98765.4321", None);
+    inner("Infinity", None);
+    inner("-Infinity", None);
+    inner("NaN", None);
+
+    // Test malformed bcd input
+    let mut good_bcd = [50; 10];
+    // Valid sign nibble
+    good_bcd[9] = 28;
+    assert!(Decimal::<13>::from_packed_bcd(&good_bcd, 0).is_ok());
+    let mut bad_bcd = [28u8; 30];
+    // Valid sign nibble, too many digits
+    bad_bcd[29] = 28;
+    assert!(Decimal::<13>::from_packed_bcd(&bad_bcd, 0).is_err());
+    // Invalid sign nibble
+    assert!(Decimal::<13>::from_packed_bcd(&[50; 10], 0).is_err());
+}
+
+#[test]
+fn decnum_round_to_place() {
+    fn inner(s: &str, expect: &[&str]) {
+        const N: usize = 13;
+        let mut cx = Context::<Decimal<N>>::default();
+        let d = cx.parse(s).unwrap();
+        for (i, expect) in expect.iter().enumerate() {
+            let mut n = d.clone();
+            cx.round_to_place(&mut n, i + 1)
+                .expect("only valid test cases");
+            assert_eq!(&n.to_string(), expect);
+        }
+    }
+    inner("599.18", &["6E+2", "6.0E+2", "599", "599.2", "599.18"]);
+    inner("99.19", &["1E+2", "99"]);
+    inner("1009", &["1E+3", "1.0E+3", "1.01E+3", "1009"]);
+}
+
+#[test]
+fn decnum_round_reduce_to_place() {
+    fn inner(s: &str, expect: &[&str]) {
+        const N: usize = 13;
+        let mut cx = Context::<Decimal<N>>::default();
+        let d = cx.parse(s).unwrap();
+        for (i, expect) in expect.iter().enumerate() {
+            let mut n = d.clone();
+            cx.round_reduce_to_place(&mut n, i + 1)
+                .expect("only valid test cases");
+            assert_eq!(&n.to_string(), expect);
+        }
+    }
+    inner("599.18", &["6E+2", "6E+2", "599", "599.2", "599.18"]);
+    inner("1009", &["1E+3", "1E+3", "1.01E+3", "1009"]);
+}
